using System;

namespace Sharp3D
{
	public class BumpMapPixelShader : IPixelShader
	{
		// t0:	texture coordinates
		// t1:	light vector in tangent space
		// t2:	half vector in tangent space
		// material.Texture0: diffuse texture
		// material.Texture1: normal map
		// material.Texture2: first horizon texture
		// material.Texture3: second horizon texture
		// material.Texture4: first trigonometry texture
		// material.Texture5: second trigonometry texture
		public Color calculate(Material material, Vector3D t0, Vector3D t1, Vector3D t2)
		{
			// diffuse color

			Color diffuseColor = material.Texture0.GetPixel(t0.X, t0.Y);


			// light and half vector

			Vector3D lightVec = t1.GetNormalized();
			Vector3D halfVec = t2.GetNormalized();


			// diffuse and specular lighting

			Color c = material.Texture1.GetPixel(t0.X, t0.Y);

			Vector3D normal = new Vector3D();
			normal.X = c.U16Scaled;
			normal.Y = -c.V16Scaled;
			normal.Z = Math.Sqrt(1.0 - normal.X*normal.X - normal.Y*normal.Y);

			Color outputColor =
				Math.Max(normal*lightVec, 0.0)*diffuseColor +
				Math.Pow(Math.Max(normal*halfVec, 0.0), 18.0)*new Color(255, 255, 255);


			// self-shadowing

			// elevations from horizon maps
			Color elev0 = material.Texture2.GetPixel(t0.X, t0.Y);
			Color elev1 = material.Texture3.GetPixel(t0.X, t0.Y);

			// angle dependent weights from trigonometry textures
			// use (lightVec.x scaled biased, lightVec.y scaled biased) as coordinates
			double coordTrigU = 0.5*lightVec.X + 0.5;
			double coordTrigV = 0.5*lightVec.Y + 0.5;

			Color weight0 = material.Texture4.GetPixel(coordTrigU, coordTrigV);
			Color weight1 = material.Texture5.GetPixel(coordTrigU, coordTrigV);

			// elevation values scaled by weights
			double elev =
				elev0.U8Scaled * weight0.RedScaled + elev0.V8Scaled * weight0.GreenScaled + elev0.W8Scaled * weight0.BlueScaled + elev0.Q8Scaled * weight0.AlphaScaled +
				elev1.U8Scaled * weight1.RedScaled + elev1.V8Scaled * weight1.GreenScaled + elev1.W8Scaled * weight1.BlueScaled + elev1.Q8Scaled * weight1.AlphaScaled;

			// elevation difference between surface and light vector scaled by a constant factor
			double shadow = 2.0 * (lightVec.Z - elev);


			return shadow * outputColor;
		}
	}
}
