using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Collections;

namespace Sharp3D {
  /// <summary>
  /// Abstraction of a graphics device. Binds a System.Windows.Forms.Control object to a DirectX
  /// surface.
  ///
  /// Usage:
  ///
  /// // Assign a canvas to the GraphicsDevice. This must be done only once during program execution,
  /// // a reference to 'myControl' is stored internally. 
  /// GraphicsDevice.Create(myControl);
  ///
  /// // For drawing, obtain a reference to the surface of the control. The surface is returned
  /// // in a locked state:
  /// Surface mySurface = GetSurface();
  ///
  /// // Now, use the surface for drawing:
  /// mySurface.Draw...();
  ///
  /// // When done drawing, the surface must be released:
  /// GraphicsDevice.ReleaseSurface();
  ///
  /// // To display the modified surface on your control, use Flip():
  /// GraphicsDevice.Flip();
  ///
  /// // Before finishing program execution, the GraphicsDevice must be destroyed:
  /// GraphicsDevice.Destroy();
  /// </summary>
  public class GraphicsDeviceException : System.ApplicationException {
    public GraphicsDeviceException(string message) : base(message) {
    }
  }
  
  public sealed class GraphicsDevice { 
    private const int maxStrLen = 200;

    private static System.Windows.Forms.Control canvas;

    /****************************************************************************
     * DLL imported methods
     ****************************************************************************/

    [System.Security.SuppressUnmanagedCodeSecurity] // We won't use this maliciously
    [DllImport("DirectXWrapper.dll", EntryPoint = "Create", CharSet=CharSet.Ansi)]
    private static extern int Create_DLL(System.IntPtr hWnd,int Width,int Height, StringBuilder errorMessage);

    [DllImport("DirectXWrapper.dll", EntryPoint = "Destroy", CharSet=CharSet.Ansi)]
    private static extern int Destroy_DLL(StringBuilder errorMessage);

    [DllImport("DirectXWrapper.dll", EntryPoint = "Lock", CharSet=CharSet.Ansi)]
    unsafe private static extern int Lock_DLL(void** pByes,int* pPitch, StringBuilder errorMessage);

    [DllImport("DirectXWrapper.dll", EntryPoint = "Unlock", CharSet=CharSet.Ansi)]
    private static extern int Unlock_DLL(StringBuilder errorMessage);

    [DllImport("DirectXWrapper.dll", EntryPoint = "Flip", CharSet=CharSet.Ansi)]
    private static extern int Flip_DLL(StringBuilder errorMessage);

    [DllImport("DirectXWrapper.dll", EntryPoint = "Resize", CharSet=CharSet.Ansi)]
    private static extern int Resize_DLL(int newWidth, int newHeight, StringBuilder errorMessage);

    [DllImport("DirectXWrapper.dll", EntryPoint = "IsInitialized")]
    private static extern int IsInitialized_DLL();

    [DllImport("DirectXWrapper.dll", EntryPoint = "IsLocked")]
    private static extern int IsLocked_DLL();

    private GraphicsDevice() { /* Private Constructor */ }

    /****************************************************************************
     * Static methods
     ****************************************************************************/
    
    public static void Resize(int newWidth, int newHeight) {
      if (IsLocked()) throw new GraphicsDeviceException("Tried to resize while surface was locked!");

      StringBuilder errorMessage = new StringBuilder(maxStrLen);
      if (Resize_DLL(newWidth, newHeight, errorMessage) != 0) throw new GraphicsDeviceException(errorMessage.ToString());
    }
    
    public static void OnCanvasResized(object sender, EventArgs args) {
      if(GraphicsDevice.IsInitialized()) {
        GraphicsDevice.Resize(canvas.Width, canvas.Height);
      }
    }
    
    public static void Create(System.Windows.Forms.Control canvas) {
    if (GraphicsDevice.canvas != null) throw new GraphicsDeviceException("Cannot create - canvas is already assigned.");

      GraphicsDevice.canvas = canvas;
      System.IntPtr hWnd = canvas.Handle;
      StringBuilder errorMessage = new StringBuilder(maxStrLen);
      if (Create_DLL(hWnd, canvas.Width, canvas.Height, errorMessage) != 0) 
      throw new GraphicsDeviceException(errorMessage.ToString());
      
      canvas.Resize += new EventHandler(GraphicsDevice.OnCanvasResized);
    }
    
    public static void Destroy() {
      if (canvas == null) return;
      StringBuilder errorMessage = new StringBuilder(maxStrLen);
      if (Destroy_DLL(errorMessage) != 0) throw new GraphicsDeviceException(errorMessage.ToString());
    }
    
    public unsafe static Surface GetSurface() {
      void* bits = null;
      int pitch;

      StringBuilder errorMessage = new StringBuilder(maxStrLen);
      int ret = Lock_DLL(&bits, &pitch, errorMessage);
      if (ret != 0) throw new GraphicsDeviceException(errorMessage.ToString());

      return new Surface((short*) bits, pitch >> 1, canvas.Width, canvas.Height);
    }

    public static void ReleaseSurface() {
      StringBuilder errorMessage = new StringBuilder(maxStrLen);
      if (Unlock_DLL(errorMessage) != 0) throw new GraphicsDeviceException(errorMessage.ToString());
    }
    
    public static void Flip() {
      StringBuilder errorMessage = new StringBuilder(maxStrLen);
      if (Flip_DLL(errorMessage) != 0) throw new GraphicsDeviceException(errorMessage.ToString());
    }

    public static bool IsInitialized() 
    {
      return (IsInitialized_DLL() != 0);
    }
    
    public static bool IsLocked() 
    {
      return (IsLocked_DLL() != 0);
    }
  }
}
