using System.Windows.Forms;
using System.Collections;
using System;
using System.IO;

namespace Sharp3D
{
	class MainFrame : Form
	{
		PictureBox canvas;

		Scene scene;

		int frame;
		bool closing;

		SceneGraphNode earth, asteroid, blast, text;

		private ArrayList GenStar(int num, Point2D midPoint, double radius)
		{
			ArrayList star = new ArrayList();

			for (int i = 0; i < num-1; i++)
			{
				Point2D start = new Point2D(midPoint.X + radius * Math.Cos(2.0*Math.PI * i/num), midPoint.Y + radius * Math.Sin(2.0*Math.PI * i/num));

				for (int k = i+1; k < num; k++)
					star.Add(new Line2D(
						start, new Point2D(midPoint.X + radius * Math.Cos(2.0*Math.PI * k/num), midPoint.Y + radius * Math.Sin(2.0*Math.PI * k/num))));
			}

			return star;
		}

		private static double CoolFunction(double x, double y) 
		{
			double d = Math.Sqrt(x * x + y * y);

			return Math.Abs(1 - 0.5 * d) * 0.4 * (Math.Cos(d * 2 * Math.PI) + 0.5);
		}

		private MainFrame()
		{
			Name = "Sharp3D";

			AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			ClientSize = new System.Drawing.Size(800, 600);

			canvas = new PictureBox();
			canvas.Dock = DockStyle.Fill;
			Controls.Add(canvas);

			Closing += new System.ComponentModel.CancelEventHandler(OnClosing);

			frame = 0;
			closing = false;


			// create scene objects

			Directory.SetCurrentDirectory("../../media");

			// earth
			earth = new SceneGraphNode();
			earth.Mesh = MeshFactory.CreateSphere(new Point3D(0.0, 0.0, 0.0), 1.5, 32, 64);
			earth.Material.Texture0 = new Texture(new Image("earth.png"), Texture.TextureModes.BILINEAR);
			earth.Material.Texture1 = new Texture(new Image("earth_norm.png"), Texture.TextureModes.NEAREST_NEIGHBOR);
			earth.Material.Texture2 = new Texture(new Image("earth_hzon0.png"), Texture.TextureModes.NEAREST_NEIGHBOR);
			earth.Material.Texture3 = new Texture(new Image("earth_hzon1.png"), Texture.TextureModes.NEAREST_NEIGHBOR);
			earth.Material.Texture4 = new Texture(new Image("earth_trig0.png"), Texture.TextureModes.NEAREST_NEIGHBOR);
			earth.Material.Texture5 = new Texture(new Image("earth_trig1.png"), Texture.TextureModes.NEAREST_NEIGHBOR);
			earth.RenderingMode = Mesh3D.RenderingModes.BUMPMAP;

			// asteroid
			asteroid = new SceneGraphNode();
			asteroid.Mesh = MeshFactory.CreateAsteroid("asteroid.dat");
			asteroid.RenderingMode = Mesh3D.RenderingModes.SPECULAR;

			// blast
			blast = new SceneGraphNode();
			blast.Mesh = MeshFactory.CreateQuadXZ(1.0, 1.0);
			blast.Material.Texture0 = new Texture(new Image("blast.png"), Texture.TextureModes.BILINEAR);
			blast.RenderingMode = Mesh3D.RenderingModes.TEXTURE;

			// text
			text = new SceneGraphNode();
			text.Mesh = MeshFactory.CreateQuadXY(new Rectangle2D(-0.6, 0.65, 0.6, 0.25));
			text.Material.Texture0 = new Texture(new Image("text.png"), Texture.TextureModes.BILINEAR);
			text.RenderingMode = Mesh3D.RenderingModes.DISPLAY;

			scene = new Scene();
			scene.Light = new PositionalLight(new Color(255, 255, 255), new Point3D(-25.0, 0.0, 0.0), 0.01);
			scene.Camera.Position = new Point3D(-6.0*Math.Cos(0.15*Math.PI), -6.0*Math.Sin(0.15*Math.PI), 0.0);
			scene.Background = new SpaceBackground(512);


			// frame output directory
			Directory.SetCurrentDirectory("movie");


			Show();
		}

		public void Loop()
		{
			while (!closing)
			{
				double time = frame / 30.0;


				// render scene

				Surface surface = GraphicsDevice.GetSurface();

				surface.Clear(scene.Fog.Color);

				// empty root node
				scene.RootNode = new SceneGraphNode();

				// earth
				earth.Transformation = Matrix.CreateRotationZ(time > 6.25f ? 0.625f : 0.1f*time);
				scene.RootNode.AddChild(earth);

				// asteroid
				Vector3D posAsteroid = (26.0/25.0*time - 8.0) * new Vector3D(1.0, 1.0, 0.0).GetNormalized();
				asteroid.Transformation = Matrix.CreateTranslation(posAsteroid) *
					Matrix.CreateRotation(
						new Vector3D(Math.Sin(0.4f*time), 2.0*Math.Cos(0.6*time), Math.Sin(0.8*time)).GetNormalized(),
						3.0 * time);
				scene.RootNode.AddChild(asteroid);

				// blast
				if (time > 6.25)
				{
					blast.Transformation =
						Matrix.CreateTranslation(-1.5*new Vector3D(1.0, 1.0, 0.0).GetNormalized()) *
						Matrix.CreateRotationZ(-0.25*Math.PI) *
						Matrix.CreateScaling(15.0*(1.0 - Math.Exp(6.25 - time)));
					scene.RootNode.AddChild(blast);
				}

				// text
				if (time > 10.0)
					scene.RootNode.AddChild(text);

				scene.Camera.Center = new Point3D(posAsteroid.X, posAsteroid.Y, posAsteroid.Z);

				scene.Render(surface, new Rectangle2D(0.0, 0.0, 800.0, 600.0));


				// save image to disk

				surface.SaveImage("frame" + frame + ".bmp");


				// prepare next frame

				GraphicsDevice.ReleaseSurface();

				GraphicsDevice.Flip();

				frame++;

				if (time > 14)
					closing = true;


				// perform Windows GUI event handling (i.e. menu handling, buttons etc.)

				Application.DoEvents();
			}
		}

		static void Main()
		{
			MainFrame frame = new MainFrame();

			// assign the frame's canvas to GraphicsDevice, i.e. all drawing 
			// operations are performed on this control
			GraphicsDevice.Create(frame.canvas);

			// start the rendering loop
			frame.Loop();

			GraphicsDevice.Destroy();
		}

		private void OnClosing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			closing = true;
		}
	}
}
