using System;

namespace Sharp3D
{
	public class MatrixException : System.ApplicationException
	{
		public MatrixException(string message) : base(message) {
		}
	}

	public class Matrix
	{
		double[,] elements = new double[4, 4];

		private bool isAffine = false;

		public Matrix()
		{
			elements[0, 0] = 1.0;
			elements[1, 0] = 0.0;
			elements[2, 0] = 0.0;
			elements[3, 0] = 0.0;

			elements[0, 1] = 0.0;
			elements[1, 1] = 1.0;
			elements[2, 1] = 0.0;
			elements[3, 1] = 0.0;

			elements[0, 2] = 0.0;
			elements[1, 2] = 0.0;
			elements[2, 2] = 1.0;
			elements[3, 2] = 0.0;

			elements[0, 3] = 0.0;
			elements[1, 3] = 0.0;
			elements[2, 3] = 0.0;
			elements[3, 3] = 1.0;

			isAffine = true;
		}

		public Matrix(Matrix matrix)
		{
			for (int i = 0; i < 4; i++)
				for (int k = 0; k < 4; k++)
					elements[i, k] = matrix.elements[i, k];

			isAffine = matrix.isAffine;
		}

		public Matrix(
			double e11, double e12, double e13, double e14,
			double e21, double e22, double e23, double e24,
			double e31, double e32, double e33, double e34,
			double e41, double e42, double e43, double e44)
		{
			elements[0, 0] = e11;
			elements[0, 1] = e12;
			elements[0, 2] = e13;
			elements[0, 3] = e14;

			elements[1, 0] = e21;
			elements[1, 1] = e22;
			elements[1, 2] = e23;
			elements[1, 3] = e24;

			elements[2, 0] = e31;
			elements[2, 1] = e32;
			elements[2, 2] = e33;
			elements[2, 3] = e34;

			elements[3, 0] = e41;
			elements[3, 1] = e42;
			elements[3, 2] = e43;
			elements[3, 3] = e44;

			isAffine = false;
		}

		public Matrix(
			double e11, double e12, double e13, double e14,
			double e21, double e22, double e23, double e24,
			double e31, double e32, double e33, double e34,
			double e41, double e42, double e43, double e44,
			bool isAffine)
		{
			elements[0, 0] = e11;
			elements[0, 1] = e12;
			elements[0, 2] = e13;
			elements[0, 3] = e14;

			elements[1, 0] = e21;
			elements[1, 1] = e22;
			elements[1, 2] = e23;
			elements[1, 3] = e24;

			elements[2, 0] = e31;
			elements[2, 1] = e32;
			elements[2, 2] = e33;
			elements[2, 3] = e34;

			elements[3, 0] = e41;
			elements[3, 1] = e42;
			elements[3, 2] = e43;
			elements[3, 3] = e44;

			this.isAffine = isAffine;
		}

		public double this[int i, int j]
		{
			get
			{
				return elements[i, j];
			}
			set
			{
				elements[i, j] = value;
			}
		}

		public bool IsAffine
		{
			get
			{
				return isAffine;
			}
		}

		public Matrix Inverted
		{
			get
			{
				if (!isAffine)
					throw new MatrixException("Inverted on non-affine matrix called.");

				Matrix T_inv = CreateTranslation(-elements[0, 3], -elements[1, 3], -elements[2, 3]);

				double s_sq_inv = 1.0 / (
					elements[0, 0]*elements[0, 0] +
					elements[1, 0]*elements[1, 0] +
					elements[2, 0]*elements[2, 0]);

				return new Matrix(
					s_sq_inv*elements[0, 0], s_sq_inv*elements[1, 0], s_sq_inv*elements[2, 0], 0.0,
					s_sq_inv*elements[0, 1], s_sq_inv*elements[1, 1], s_sq_inv*elements[2, 1], 0.0,
					s_sq_inv*elements[0, 2], s_sq_inv*elements[1, 2], s_sq_inv*elements[2, 2], 0.0,
					0.0, 0.0, 0.0, 1.0, true) * T_inv;
			}
		}

		static public Matrix operator* (Matrix matrixL, Matrix matrixR)
		{
			Matrix ret = new Matrix();

			for (int i = 0; i < 4; i++)
				for (int j = 0; j < 4; j++)
					ret[i, j] =
						matrixL[i, 0]*matrixR[0, j] +
						matrixL[i, 1]*matrixR[1, j] +
						matrixL[i, 2]*matrixR[2, j] +
						matrixL[i, 3]*matrixR[3, j];

			ret.isAffine = matrixL.isAffine && matrixR.isAffine;

			return ret;
		}

		static public Vector3D operator* (Matrix matrix, Vector3D vector)
		{
			if (!matrix.isAffine)
				throw new MatrixException("Can't apply non-affine matrix operation to vector.");

			return new Vector3D(
				matrix[0, 0]*vector.X + matrix[0, 1]*vector.Y + matrix[0, 2]*vector.Z,
				matrix[1, 0]*vector.X + matrix[1, 1]*vector.Y + matrix[1, 2]*vector.Z,
				matrix[2, 0]*vector.X + matrix[2, 1]*vector.Y + matrix[2, 2]*vector.Z);
		}

		static public Point3D operator* (Matrix matrix, Point3D point)
		{
			double W = matrix[3, 0]*point.X + matrix[3, 1]*point.Y + matrix[3, 2]*point.Z + matrix[3, 3];

			if (W == 0.0)
				throw new MatrixException("Homogeneous W zero when transforming point.");

			return new Point3D(
				(matrix[0, 0]*point.X + matrix[0, 1]*point.Y + matrix[0, 2]*point.Z + matrix[0, 3]) / W,
				(matrix[1, 0]*point.X + matrix[1, 1]*point.Y + matrix[1, 2]*point.Z + matrix[1, 3]) / W,
				(matrix[2, 0]*point.X + matrix[2, 1]*point.Y + matrix[2, 2]*point.Z + matrix[2, 3]) / W);
		}

		static public Matrix CreateNullMatrix()
		{
			return new Matrix(
				0.0, 0.0, 0.0, 0.0,
				0.0, 0.0, 0.0, 0.0,
				0.0, 0.0, 0.0, 0.0,
				0.0, 0.0, 0.0, 0.0,
				false);
		}

		static public Matrix CreateIdentityMatrix()
		{
			return new Matrix();
		}

		static public Matrix CreateTranslation(double tX, double tY, double tZ)
		{
			return new Matrix(
				1.0, 0.0, 0.0, tX,
				0.0, 1.0, 0.0, tY,
				0.0, 0.0, 1.0, tZ,
				0.0, 0.0, 0.0, 1.0,
				true);
		}

		static public Matrix CreateTranslation(Vector3D vector)
		{
			return new Matrix(
				1.0, 0.0, 0.0, vector.X,
				0.0, 1.0, 0.0, vector.Y,
				0.0, 0.0, 1.0, vector.Z,
				0.0, 0.0, 0.0, 1.0,
				true);
		}

		static public Matrix CreateScaling(double s)
		{
			return new Matrix(
				s, 0.0, 0.0, 0.0,
				0.0, s, 0.0, 0.0,
				0.0, 0.0, s, 0.0,
				0.0, 0.0, 0.0, 1.0,
				true);
		}

		static public Matrix CreateRotationX(double phi)
		{
			double sin_phi = Math.Sin(phi);
			double cos_phi = Math.Cos(phi);

			return new Matrix(
				1.0, 0.0, 0.0, 0.0,
				0.0, cos_phi, -sin_phi, 0.0,
				0.0, sin_phi, cos_phi, 0.0,
				0.0, 0.0, 0.0, 1.0,
				true);
		}

		static public Matrix CreateRotationY(double phi)
		{
			double sin_phi = Math.Sin(phi);
			double cos_phi = Math.Cos(phi);

			return new Matrix(
				cos_phi, 0.0, -sin_phi, 0.0,
				0.0, 1.0, 0.0, 0.0,
				sin_phi, 0.0, cos_phi, 0.0,
				0.0, 0.0, 0.0, 1.0,
				true);
		}

		static public Matrix CreateRotationZ(double phi)
		{
			double sin_phi = Math.Sin(phi);
			double cos_phi = Math.Cos(phi);

			return new Matrix(
				cos_phi, -sin_phi, 0.0, 0.0,
				sin_phi, cos_phi, 0.0, 0.0,
				0.0, 0.0, 1.0, 0.0,
				0.0, 0.0, 0.0, 1.0,
				true);
		}

		static public Matrix CreateRotation(Vector3D axis, double phi)
		{
			axis.Normalize();
			Vector3D b = axis.CreateOrthogonalVector();
			Vector3D c = axis % b;

			Matrix Ma_x = new Matrix(
				axis.X, axis.Y, axis.Z, 0.0,
				b.X, b.Y, b.Z, 0.0,
				c.X, c.Y, c.Z, 0.0,
				0.0, 0.0, 0.0, 1.0, true);

			Matrix Mx_a = new Matrix(
				axis.X, b.X, c.X, 0.0,
				axis.Y, b.Y, c.Y, 0.0,
				axis.Z, b.Z, c.Z, 0.0,
				0.0, 0.0, 0.0, 1.0, true);

			return Mx_a * CreateRotationX(phi) * Ma_x;
		}

		static public Matrix CreateRotation(Point3D p, Vector3D axis, double phi)
		{
			return CreateTranslation(p.X, p.Y, p.Z) * CreateRotation(axis, phi) * CreateTranslation(-p.X, -p.Y, -p.Z);
		}

		static public Matrix CreateTransformationFromTo(
			Point3D p0, Vector3D v0, Vector3D u0,
			Point3D p1, Vector3D v1, Vector3D u1)
		{
			u0 = (v0 % u0) % v0;
			u1 = (v1 % u1) % v1;

			u0.Normalize();
			v0.Normalize();
			u1.Normalize();
			v1.Normalize();

			Vector3D w0 = u0 % v0;
			Vector3D w1 = u1 % v1;

			Matrix MCS0_xyzO = new Matrix(
				u0.X, u0.Y, u0.Z, - u0.X*p0.X - u0.Y*p0.Y - u0.Z*p0.Z,
				v0.X, v0.Y, v0.Z, - v0.X*p0.X - v0.Y*p0.Y - v0.Z*p0.Z,
				w0.X, w0.Y, w0.Z, - w0.X*p0.X - w0.Y*p0.Y - w0.Z*p0.Z,
				0.0, 0.0, 0.0, 1.0, true);

			Matrix MxyzO_CS1 = new Matrix(
				u1.X, v1.X, w1.X, p1.X,
				u1.Y, v1.Y, w1.Y, p1.Y,
				u1.Z, v1.Z, w1.Z, p1.Z,
				0.0, 0.0, 0.0, 1.0, true);

			return MxyzO_CS1 * MCS0_xyzO;
		}
	}
}
