using System;
using System.Collections;

namespace Sharp3D
{
	public class Mesh3D
	{
		private ArrayList triangles;

		public enum RenderingModes
		{
			WIREFRAME,
			FLAT,
			GOURAUD,
			BUMPMAP,
			SPECULAR,
			TEXTURE,
			DISPLAY
		};

		public Mesh3D()
		{
			triangles = new ArrayList();
		}

		public void AddTriangle(Vertex v1, Vertex v2, Vertex v3)
		{
			triangles.Add(new Polygon3D(v1, v2, v3));
		}

		public int TriangleCount
		{
			get
			{
				return triangles.Count;
			}
		}

		public Polygon3D GetTriangle(int i)
		{
			return (Polygon3D) triangles[i];
		}

		public void RenderWireframe(RenderingContext context, Color color)
		{
			Clipper3D clipper = new Clipper3D(context.Camera.NearPlaneDistance, context.Camera.FarPlaneDistance);

			Matrix matrixLV = context.Camera2Canonical * context.World2Camera * context.Local2World;
			Matrix matrixVD = context.Screen2Device * context.Canonical2Screen;

			if (context.BackfaceCulling)
			{
				foreach (Polygon3D polygon in triangles)
				{
					Polygon3D transformed = new Polygon3D();

					transformed.Add(new Vertex(matrixLV * polygon[0].Position));
					transformed.Add(new Vertex(matrixLV * polygon[1].Position));
					transformed.Add(new Vertex(matrixLV * polygon[2].Position));

					Vector3D n = (transformed[1].Position - transformed[0].Position) % (transformed[2].Position - transformed[0].Position);
					if ((transformed[0].Position - new Point3D(0.0, 0.0, 0.0)) * n < 0.0)
					{
						Polygon3D clipped;
						if (clipper.Clip(transformed, out clipped))
						{
							for (int i = 0; i < clipped.Size; i++)
								clipped[i].Position = matrixVD * clipped[i].Position;

							for (int i = 0; i < clipped.Size; i++)
							{
								context.Surface.DrawLineBresenham(new Line2D(
									new Point2D(clipped[i].Position.X, clipped[i].Position.Y),
									new Point2D(clipped[(i+1)%clipped.Size].Position.X, clipped[(i+1)%clipped.Size].Position.Y)), color);
							}
						}
					}
				}
			}
			else
			{
				foreach (Polygon3D polygon in triangles)
				{
					Polygon3D transformed = new Polygon3D();

					transformed.Add(new Vertex(matrixLV * polygon[0].Position));
					transformed.Add(new Vertex(matrixLV * polygon[1].Position));
					transformed.Add(new Vertex(matrixLV * polygon[2].Position));

					Polygon3D clipped;
					if (clipper.Clip(transformed, out clipped))
					{
						for (int i = 0; i < clipped.Size; i++)
							clipped[i].Position = matrixVD * clipped[i].Position;

						for (int i = 0; i < clipped.Size; i++)
						{
							context.Surface.DrawLineBresenham(new Line2D(
								new Point2D(clipped[i].Position.X, clipped[i].Position.Y),
								new Point2D(clipped[(i+1)%clipped.Size].Position.X, clipped[(i+1)%clipped.Size].Position.Y)), color);
						}
					}
				}
			}
		}

		public void RenderFlat(RenderingContext context, Material material)
		{
			Clipper3D clipper = new Clipper3D(context.Camera.NearPlaneDistance, context.Camera.FarPlaneDistance);

			Matrix matrixLV = context.Camera2Canonical * context.World2Camera * context.Local2World;
			Matrix matrixVD = context.Screen2Device * context.Canonical2Screen;

			ILight lightLCS = context.Light.GetTransformed(context.Local2World.Inverted);

			foreach (Polygon3D polygon in triangles)
			{
				Polygon3D transformed = new Polygon3D();

				transformed.Add(new Vertex(matrixLV * polygon[0].Position));
				transformed.Add(new Vertex(matrixLV * polygon[1].Position));
				transformed.Add(new Vertex(matrixLV * polygon[2].Position));

				Vector3D n = (transformed[1].Position - transformed[0].Position) % (transformed[2].Position - transformed[0].Position);
				if ((transformed[0].Position - new Point3D(0.0, 0.0, 0.0)) * n < 0.0)
				{
					Polygon3D clipped;
					if (clipper.Clip(transformed, out clipped))
					{
						// transform to device coordinates

						for (int i = 0; i < clipped.Size; i++)
							clipped[i].Position = matrixVD * clipped[i].Position;

						// lighting calculations in local coordinate system

						Point3D midPoint = new Point3D(
							(polygon[0].Position.X + polygon[1].Position.X + polygon[2].Position.X) / 3.0,
							(polygon[0].Position.Y + polygon[1].Position.Y + polygon[2].Position.Y) / 3.0,
							(polygon[0].Position.Z + polygon[1].Position.Z + polygon[2].Position.Z) / 3.0);

						Vector3D normal = ((polygon[1].Position - polygon[0].Position) % (polygon[2].Position - polygon[0].Position)).GetNormalized();

						Color colorPixel =
							material.AmbientCoefficient * (context.AmbientColor * material.Color) +
							(material.DiffuseCoefficient * lightLCS.CalcDiffuseWeight(midPoint, normal, material)) * (lightLCS.Color * material.Color) +
							(material.SpecularCoefficient * lightLCS.CalcSpecularWeight(midPoint, normal, context.Camera.Position, material)) * lightLCS.Color;

						RenderFlatPolygon(clipped, context, colorPixel);
					}
				}
			}
		}

		private class EdgeFlat : IComparable
		{
			public int ymin, ymax;
			public double x, dx;
			public double z, dz;

			public EdgeFlat(int ymin, int ymax, double x, double dx, double z, double dz)
			{
				this.ymin = ymin;
				this.ymax = ymax;

				this.x = x;
				this.dx = dx;

				this.z = z;
				this.dz = dz;
			}

			public int CompareTo(object obj)
			{
				if (x < ((EdgeFlat)obj).x)
				{
					return -1;
				}
				else if (x == ((EdgeFlat)obj).x)
				{
					return 0;
				}
				else
				{
					return 1;
				}
			}
		}

		private void RenderFlatPolygon(Polygon3D polygon, RenderingContext context, Color color)
		{
			// create edge table

			ArrayList ET = new ArrayList();
			ArrayList AET = new ArrayList();

			double ymin = context.Viewport.Bottom;

			for (int i = 0; i < polygon.Size; i++)
			{
				double y0 = polygon[i].Position.Y;
				double y1 = polygon[(i+1)%polygon.Size].Position.Y;

				if (y0 != y1)
				{
					double x0 = polygon[i].Position.X;
					double x1 = polygon[(i+1)%polygon.Size].Position.X;

					double z0 = polygon[i].Position.Z;
					double z1 = polygon[(i+1)%polygon.Size].Position.Z;

					if (y0 < y1)
					{
						ET.Add(new EdgeFlat((int)y0, (int)y1, x0, (x1 - x0)/(y1 - y0), z0, (z1 - z0)/(y1 - y0)));

						if (y0 < ymin)
							ymin = y0;
					}
					else
					{
						ET.Add(new EdgeFlat((int)y1, (int)y0, x1, (x1 - x0)/(y1 - y0), z1, (z1 - z0)/(y1 - y0)));

						if (y1 < ymin)
							ymin = y1;
					}
				}
			}

			// draw pixels

			int y = (int)ymin;
			while (ET.Count != 0 || AET.Count != 0)
			{
				int i = 0;
				while (i != ET.Count)
				{
					if (((EdgeFlat)ET[i]).ymin == y)
					{
						AET.Add(ET[i]);
						ET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				i = 0;
				while (i != AET.Count)
				{
					if (((EdgeFlat)AET[i]).ymax == y)
					{
						AET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				if (AET.Count % 2 == 1)
					throw new ApplicationException("must be even number of corners");

				AET.Sort();
				for (i = 0; i < AET.Count/2; i++)
				{
					EdgeFlat e0 = (EdgeFlat)AET[2*i];
					EdgeFlat e1 = (EdgeFlat)AET[2*i+1];

					double deltaz = (e1.z - e0.z)/(e1.x - e0.x);
					double z = 65535.0 - e0.z;

					// rounding down when casting to int
					for (int x = (int)e0.x; x < (int)e1.x; x++)
					{
						if (z > context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)])
						{
							context.Surface.SetPixelAlpha(x, y, context.Fog.applyFog(1.0 - z/65535.0, color));
							context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)] = (ushort)z;
						}

						z -= deltaz;
					}

					e0.x += e0.dx;
					e1.x += e1.dx;
					e0.z += e0.dz;
					e1.z += e1.dz;
				}

				y++;
			}
		}

		public void RenderGouraud(RenderingContext context, Material material)
		{
			Clipper3D clipper = new Clipper3D(context.Camera.NearPlaneDistance, context.Camera.FarPlaneDistance);

			Matrix matrixLV = context.Camera2Canonical * context.World2Camera * context.Local2World;
			Matrix matrixVD = context.Screen2Device * context.Canonical2Screen;

			ILight lightLCS = context.Light.GetTransformed(context.Local2World.Inverted);

			foreach (Polygon3D polygon in triangles)
			{
				Polygon3D transformed = new Polygon3D();

				// don't need normal
				transformed.Add(new Vertex(matrixLV * polygon[0].Position, polygon[0].TextureCoordinate0));
				transformed.Add(new Vertex(matrixLV * polygon[1].Position, polygon[1].TextureCoordinate0));
				transformed.Add(new Vertex(matrixLV * polygon[2].Position, polygon[2].TextureCoordinate0));

				// backface culling
				Vector3D n = (transformed[1].Position - transformed[0].Position) % (transformed[2].Position - transformed[0].Position);
				if ((transformed[0].Position - new Point3D(0.0, 0.0, 0.0)) * n < 0.0)
				{
					// lighting calculations in local coordinate system
					for (int i = 0; i < 3; i++)
					{
						transformed[i].DiffuseWeight = lightLCS.CalcDiffuseWeight(polygon[i].Position, polygon[i].Normal, material);
						transformed[i].SpecularWeight = lightLCS.CalcSpecularWeight(polygon[i].Position, polygon[i].Normal, context.Camera.Position, material);
					}

					Polygon3D clipped;
					if (clipper.Clip(transformed, out clipped))
					{
						// transform to device coordinates

						for (int i = 0; i < clipped.Size; i++)
							clipped[i].Position = matrixVD * clipped[i].Position;

						if (material.Texture0 == null)
							RenderGouraudPolygon(clipped, context, material);
						else
							RenderGouraudTexturePolygon(clipped, context, material);
					}
				}
			}
		}

		private class EdgeGouraud : IComparable
		{
			public int ymin, ymax;
			public double x, dx;
			public double z, dz;
			public double D, dD;
			public double S, dS;

			public EdgeGouraud(int ymin, int ymax, double x, double dx, double z, double dz, double D, double dD, double S, double dS)
			{
				this.ymin = ymin;
				this.ymax = ymax;

				this.x = x;
				this.dx = dx;

				this.z = z;
				this.dz = dz;

				this.D = D;
				this.dD = dD;

				this.S = S;
				this.dS = dS;
			}

			public int CompareTo(object obj)
			{
				if (x < ((EdgeGouraud)obj).x)
				{
					return -1;
				}
				else if (x == ((EdgeGouraud)obj).x)
				{
					return 0;
				}
				else
				{
					return 1;
				}
			}
		}

		private void RenderGouraudPolygon(Polygon3D polygon, RenderingContext context, Material material)
		{
			// create edge table

			ArrayList ET = new ArrayList();
			ArrayList AET = new ArrayList();

			double ymin = context.Viewport.Bottom;

			for (int i = 0; i < polygon.Size; i++)
			{
				double y0 = polygon[i].Position.Y;
				double y1 = polygon[(i+1)%polygon.Size].Position.Y;

				if (y0 != y1)
				{
					double x0 = polygon[i].Position.X;
					double x1 = polygon[(i+1)%polygon.Size].Position.X;

					double z0 = polygon[i].Position.Z;
					double z1 = polygon[(i+1)%polygon.Size].Position.Z;

					double D0 = polygon[i].DiffuseWeight;
					double D1 = polygon[(i+1)%polygon.Size].DiffuseWeight;

					double S0 = polygon[i].SpecularWeight;
					double S1 = polygon[(i+1)%polygon.Size].SpecularWeight;

					if (y0 < y1)
					{
						ET.Add(new EdgeGouraud((int)y0, (int)y1, x0, (x1 - x0)/(y1 - y0), z0, (z1 - z0)/(y1 - y0), D0, (D1 - D0)/(y1 - y0), S0, (S1 - S0)/(y1 - y0)));

						if (y0 < ymin)
							ymin = y0;
					}
					else
					{
						ET.Add(new EdgeGouraud((int)y1, (int)y0, x1, (x1 - x0)/(y1 - y0), z1, (z1 - z0)/(y1 - y0), D1, (D1 - D0)/(y1 - y0), S1, (S1 - S0)/(y1 - y0)));

						if (y1 < ymin)
							ymin = y1;
					}
				}
			}

			// draw pixels

			int y = (int)ymin;
			while (ET.Count != 0 || AET.Count != 0)
			{
				int i = 0;
				while (i != ET.Count)
				{
					if (((EdgeGouraud)ET[i]).ymin == y)
					{
						AET.Add(ET[i]);
						ET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				i = 0;
				while (i != AET.Count)
				{
					if (((EdgeGouraud)AET[i]).ymax == y)
					{
						AET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				if (AET.Count % 2 == 1)
					throw new ApplicationException("must be even number of corners");

				AET.Sort();
				for (i = 0; i < AET.Count/2; i++)
				{
					EdgeGouraud e0 = (EdgeGouraud)AET[2*i];
					EdgeGouraud e1 = (EdgeGouraud)AET[2*i+1];

					double deltaz = (e1.z - e0.z)/(e1.x - e0.x);
					double z = 65535.0 - e0.z;

					double deltaD = (e1.D - e0.D)/(e1.x - e0.x);
					double D = e0.D;

					double deltaS = (e1.S - e0.S)/(e1.x - e0.x);
					double S = e0.S;

					// rounding down when casting to int
					for (int x = (int)e0.x; x < (int)e1.x; x++)
					{
						if (z > context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)])
						{
							context.Surface.SetPixelAlpha(x, y,
								context.Fog.applyFog(1.0 - z/65535.0,
									material.AmbientCoefficient * (context.AmbientColor * material.Color) +
									(material.DiffuseCoefficient * D) * (context.Light.Color * material.Color) +
									(material.SpecularCoefficient * S) * context.Light.Color));

							context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)] = (ushort)z;
						}

						z -= deltaz;
						D += deltaD;
						S += deltaS;
					}

					e0.x += e0.dx;
					e1.x += e1.dx;

					e0.z += e0.dz;
					e1.z += e1.dz;

					e0.D += e0.dD;
					e1.D += e1.dD;

					e0.S += e0.dS;
					e1.S += e1.dS;
				}

				y++;
			}
		}

		private class EdgeGouraudTexture : IComparable
		{
			public int ymin, ymax;
			public double x, dx;
			public double z, dz;
			public double D, dD;
			public double S, dS;
			public double um, dum;
			public double vm, dvm;

			public EdgeGouraudTexture(
				int ymin, int ymax, double x, double dx, double z, double dz,
				double D, double dD, double S, double dS,
				double um, double dum, double vm, double dvm)
			{
				this.ymin = ymin;
				this.ymax = ymax;

				this.x = x;
				this.dx = dx;

				this.z = z;
				this.dz = dz;

				this.D = D;
				this.dD = dD;

				this.S = S;
				this.dS = dS;

				this.um = um;
				this.dum = dum;

				this.vm = vm;
				this.dvm = dvm;
			}

			public int CompareTo(object obj)
			{
				if (x < ((EdgeGouraudTexture)obj).x)
				{
					return -1;
				}
				else if (x == ((EdgeGouraudTexture)obj).x)
				{
					return 0;
				}
				else
				{
					return 1;
				}
			}
		}

		private void RenderGouraudTexturePolygon(Polygon3D polygon, RenderingContext context, Material material)
		{
			// create edge table

			ArrayList ET = new ArrayList();
			ArrayList AET = new ArrayList();

			double ymin = context.Viewport.Bottom;

			for (int i = 0; i < polygon.Size; i++)
			{
				double y0 = polygon[i].Position.Y;
				double y1 = polygon[(i+1)%polygon.Size].Position.Y;

				if (y0 != y1)
				{
					double x0 = polygon[i].Position.X;
					double x1 = polygon[(i+1)%polygon.Size].Position.X;

					double z0 = polygon[i].Position.Z;
					double z1 = polygon[(i+1)%polygon.Size].Position.Z;

					double D0 = polygon[i].DiffuseWeight;
					double D1 = polygon[(i+1)%polygon.Size].DiffuseWeight;

					double S0 = polygon[i].SpecularWeight;
					double S1 = polygon[(i+1)%polygon.Size].SpecularWeight;

					double um0 = polygon[i].TextureCoordinate0.X * (z0 - context.PerspectiveCorrection);
					double um1 = polygon[(i+1)%polygon.Size].TextureCoordinate0.X * (z1 - context.PerspectiveCorrection);

					double vm0 = polygon[i].TextureCoordinate0.Y * (z0 - context.PerspectiveCorrection);
					double vm1 = polygon[(i+1)%polygon.Size].TextureCoordinate0.Y * (z1 - context.PerspectiveCorrection);

					double dy = y1 - y0;

					if (y0 < y1)
					{
						ET.Add(new EdgeGouraudTexture((int)y0, (int)y1, x0, (x1 - x0)/dy, z0, (z1 - z0)/dy, D0, (D1 - D0)/dy, S0, (S1 - S0)/dy, um0, (um1 - um0)/dy, vm0, (vm1 - vm0)/dy));

						if (y0 < ymin)
							ymin = y0;
					}
					else
					{
						ET.Add(new EdgeGouraudTexture((int)y1, (int)y0, x1, (x1 - x0)/dy, z1, (z1 - z0)/dy, D1, (D1 - D0)/dy, S1, (S1 - S0)/dy, um1, (um1 - um0)/dy, vm1, (vm1 - vm0)/dy));

						if (y1 < ymin)
							ymin = y1;
					}
				}
			}

			// draw pixels

			int y = (int)ymin;
			while (ET.Count != 0 || AET.Count != 0)
			{
				int i = 0;
				while (i != ET.Count)
				{
					if (((EdgeGouraudTexture)ET[i]).ymin == y)
					{
						AET.Add(ET[i]);
						ET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				i = 0;
				while (i != AET.Count)
				{
					if (((EdgeGouraudTexture)AET[i]).ymax == y)
					{
						AET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				if (AET.Count % 2 == 1)
					throw new ApplicationException("must be even number of corners");

				AET.Sort();
				for (i = 0; i < AET.Count/2; i++)
				{
					EdgeGouraudTexture e0 = (EdgeGouraudTexture)AET[2*i];
					EdgeGouraudTexture e1 = (EdgeGouraudTexture)AET[2*i+1];

					double ddZ = (e1.z - e0.z)/(e1.x - e0.x);
					double z = e0.z;

					double ddD = (e1.D - e0.D)/(e1.x - e0.x);
					double D = e0.D;

					double ddS = (e1.S - e0.S)/(e1.x - e0.x);
					double S = e0.S;

					double ddum = (e1.um - e0.um)/(e1.x - e0.x);
					double um = e0.um;

					double ddvm = (e1.vm - e0.vm)/(e1.x - e0.x);
					double vm = e0.vm;

					// rounding down when casting to int
					for (int x = (int)e0.x; x < (int)e1.x; x++)
					{
						if (65535.0 - z > context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)])
						{
							Color color = material.Texture0.GetPixel(
								um/(z - context.PerspectiveCorrection),
								vm/(z - context.PerspectiveCorrection));

							context.Surface.SetPixelAlpha(x, y,
								context.Fog.applyFog(z/65535.0,
								material.AmbientCoefficient * (context.AmbientColor * color) +
								(material.DiffuseCoefficient * D) * (context.Light.Color * color) +
								(material.SpecularCoefficient * S) * context.Light.Color));

							context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)] = (ushort)(65535.0 - z);
						}

						z += ddZ;
						D += ddD;
						S += ddS;
						um += ddum;
						vm += ddvm;
					}

					e0.x += e0.dx;
					e1.x += e1.dx;

					e0.z += e0.dz;
					e1.z += e1.dz;

					e0.D += e0.dD;
					e1.D += e1.dD;

					e0.S += e0.dS;
					e1.S += e1.dS;

					e0.um += e0.dum;
					e1.um += e1.dum;

					e0.vm += e0.dvm;
					e1.vm += e1.dvm;
				}

				y++;
			}
		}

		public void RenderShader(RenderingContext context, Material material, IVertexShader vertexShader, IPixelShader pixelShader)
		{
			Matrix matrixVD = context.Screen2Device * context.Canonical2Screen;

			Clipper3D clipper = new Clipper3D(context.Camera.NearPlaneDistance, context.Camera.FarPlaneDistance);

			foreach (Polygon3D polygon in triangles)
			{
				Polygon3D transformed = new Polygon3D();

				transformed.Add(vertexShader.calculate(polygon[0]));
				transformed.Add(vertexShader.calculate(polygon[1]));
				transformed.Add(vertexShader.calculate(polygon[2]));

				// backface culling
				Vector3D n = (transformed[1].Position - transformed[0].Position) % (transformed[2].Position - transformed[0].Position);
				if ((transformed[0].Position - new Point3D(0.0, 0.0, 0.0)) * n < 0.0)
				{
					Polygon3D clipped;
					if (clipper.Clip(transformed, out clipped))
					{
						// transform to device coordinates

						for (int i = 0; i < clipped.Size; i++)
							clipped[i].Position = matrixVD * clipped[i].Position;

						RenderShaderPolygon(clipped, context, material, pixelShader);
					}
				}
			}
		}

		// assume position is in screen coordinate system
		public void RenderDisplay(RenderingContext context, Material material, IPixelShader pixelShader)
		{
			foreach (Polygon3D polygon in triangles)
			{
				Polygon3D devicePolygon = new Polygon3D();

				for (int i = 0; i < 3; i++)
				{
					devicePolygon.Add(new Vertex(
						context.Screen2Device * polygon[i].Position,
						polygon[i].Normal,
						polygon[i].Tangent,
						polygon[i].DiffuseWeight,
						polygon[i].SpecularWeight,
						polygon[i].TextureCoordinate0,
						polygon[i].TextureCoordinate1,
						polygon[i].TextureCoordinate2));
				}

				RenderShaderPolygon(devicePolygon, context, material, pixelShader);
			}
		}

		private class EdgeShader : IComparable
		{
			public int ymin, ymax;
			public double x, dx;
			public double z, dz;

			public double um0, dum0;
			public double vm0, dvm0;
			public double wm0, dwm0;

			public double um1, dum1;
			public double vm1, dvm1;
			public double wm1, dwm1;

			public double um2, dum2;
			public double vm2, dvm2;
			public double wm2, dwm2;

			public EdgeShader(
				int ymin, int ymax, double x, double dx, double z, double dz,
				double um0, double dum0, double vm0, double dvm0, double wm0, double dwm0,
				double um1, double dum1, double vm1, double dvm1, double wm1, double dwm1,
				double um2, double dum2, double vm2, double dvm2, double wm2, double dwm2)
			{
				this.ymin = ymin;
				this.ymax = ymax;
				this.x = x;
				this.dx = dx;
				this.z = z;
				this.dz = dz;

				this.um0 = um0;
				this.dum0 = dum0;
				this.vm0 = vm0;
				this.dvm0 = dvm0;
				this.wm0 = wm0;
				this.dwm0 = dwm0;

				this.um1 = um1;
				this.dum1 = dum1;
				this.vm1 = vm1;
				this.dvm1 = dvm1;
				this.wm1 = wm1;
				this.dwm1 = dwm1;

				this.um2 = um2;
				this.dum2 = dum2;
				this.vm2 = vm2;
				this.dvm2 = dvm2;
				this.wm2 = wm2;
				this.dwm2 = dwm2;
			}

			public int CompareTo(object obj)
			{
				if (x < ((EdgeShader)obj).x)
				{
					return -1;
				}
				else if (x == ((EdgeShader)obj).x)
				{
					return 0;
				}
				else
				{
					return 1;
				}
			}
		}

		private void RenderShaderPolygon(Polygon3D polygon, RenderingContext context, Material material, IPixelShader shader)
		{
			// create edge table

			ArrayList ET = new ArrayList();
			ArrayList AET = new ArrayList();

			double ymin = context.Viewport.Bottom;

			for (int i = 0; i < polygon.Size; i++)
			{
				double yS = polygon[i].Position.Y;
				double yE = polygon[(i+1)%polygon.Size].Position.Y;

				if (yS != yE)
				{
					double xS = polygon[i].Position.X;
					double xE = polygon[(i+1)%polygon.Size].Position.X;

					double zS = polygon[i].Position.Z;
					double zE = polygon[(i+1)%polygon.Size].Position.Z;

					double um0S = polygon[i].TextureCoordinate0.X * (zS - context.PerspectiveCorrection);
					double um0E = polygon[(i+1)%polygon.Size].TextureCoordinate0.X * (zE - context.PerspectiveCorrection);
					double vm0S = polygon[i].TextureCoordinate0.Y * (zS - context.PerspectiveCorrection);
					double vm0E = polygon[(i+1)%polygon.Size].TextureCoordinate0.Y * (zE - context.PerspectiveCorrection);
					double wm0S = polygon[i].TextureCoordinate0.Z * (zS - context.PerspectiveCorrection);
					double wm0E = polygon[(i+1)%polygon.Size].TextureCoordinate0.Z * (zE - context.PerspectiveCorrection);

					double um1S = polygon[i].TextureCoordinate1.X * (zS - context.PerspectiveCorrection);
					double um1E = polygon[(i+1)%polygon.Size].TextureCoordinate1.X * (zE - context.PerspectiveCorrection);
					double vm1S = polygon[i].TextureCoordinate1.Y * (zS - context.PerspectiveCorrection);
					double vm1E = polygon[(i+1)%polygon.Size].TextureCoordinate1.Y * (zE - context.PerspectiveCorrection);
					double wm1S = polygon[i].TextureCoordinate1.Z * (zS - context.PerspectiveCorrection);
					double wm1E = polygon[(i+1)%polygon.Size].TextureCoordinate1.Z * (zE - context.PerspectiveCorrection);

					double um2S = polygon[i].TextureCoordinate2.X * (zS - context.PerspectiveCorrection);
					double um2E = polygon[(i+1)%polygon.Size].TextureCoordinate2.X * (zE - context.PerspectiveCorrection);
					double vm2S = polygon[i].TextureCoordinate2.Y * (zS - context.PerspectiveCorrection);
					double vm2E = polygon[(i+1)%polygon.Size].TextureCoordinate2.Y * (zE - context.PerspectiveCorrection);
					double wm2S = polygon[i].TextureCoordinate2.Z * (zS - context.PerspectiveCorrection);
					double wm2E = polygon[(i+1)%polygon.Size].TextureCoordinate2.Z * (zE - context.PerspectiveCorrection);

					double dy = yE - yS;

					if (yS < yE)
					{
						ET.Add(new EdgeShader(
							(int)yS, (int)yE, xS, (xE - xS)/dy, zS, (zE - zS)/dy,
							um0S, (um0E - um0S)/dy, vm0S, (vm0E - vm0S)/dy, wm0S, (wm0E - wm0S)/dy,
							um1S, (um1E - um1S)/dy, vm1S, (vm1E - vm1S)/dy, wm1S, (wm1E - wm1S)/dy,
							um2S, (um2E - um2S)/dy, vm2S, (vm2E - vm2S)/dy, wm2S, (wm2E - wm2S)/dy));

						if (yS < ymin)
							ymin = yS;
					}
					else
					{
						ET.Add(new EdgeShader(
							(int)yE, (int)yS, xE, (xE - xS)/dy, zE, (zE - zS)/dy,
							um0E, (um0E - um0S)/dy, vm0E, (vm0E - vm0S)/dy, wm0E, (wm0E - wm0S)/dy,
							um1E, (um1E - um1S)/dy, vm1E, (vm1E - vm1S)/dy, wm1E, (wm1E - wm1S)/dy,
							um2E, (um2E - um2S)/dy, vm2E, (vm2E - vm2S)/dy, wm2E, (wm2E - wm2S)/dy));

						if (yE < ymin)
							ymin = yE;
					}
				}
			}

			// draw pixels

			int y = (int)ymin;
			while (ET.Count != 0 || AET.Count != 0)
			{
				int i = 0;
				while (i != ET.Count)
				{
					if (((EdgeShader)ET[i]).ymin == y)
					{
						AET.Add(ET[i]);
						ET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				i = 0;
				while (i != AET.Count)
				{
					if (((EdgeShader)AET[i]).ymax == y)
					{
						AET.RemoveAt(i);
					}
					else
					{
						i++;
					}
				}

				if (AET.Count % 2 == 1)
					throw new ApplicationException("must be even number of corners");

				AET.Sort();
				for (i = 0; i < AET.Count/2; i++)
				{
					EdgeShader e0 = (EdgeShader)AET[2*i];
					EdgeShader e1 = (EdgeShader)AET[2*i+1];

					double ddZ = (e1.z - e0.z)/(e1.x - e0.x);
					double z = e0.z;

					double ddum0 = (e1.um0 - e0.um0)/(e1.x - e0.x);
					double um0 = e0.um0;
					double ddvm0 = (e1.vm0 - e0.vm0)/(e1.x - e0.x);
					double vm0 = e0.vm0;
					double ddwm0 = (e1.wm0 - e0.wm0)/(e1.x - e0.x);
					double wm0 = e0.wm0;

					double ddum1 = (e1.um1 - e0.um1)/(e1.x - e0.x);
					double um1 = e0.um1;
					double ddvm1 = (e1.vm1 - e0.vm1)/(e1.x - e0.x);
					double vm1 = e0.vm1;
					double ddwm1 = (e1.wm1 - e0.wm1)/(e1.x - e0.x);
					double wm1 = e0.wm1;

					double ddum2 = (e1.um2 - e0.um2)/(e1.x - e0.x);
					double um2 = e0.um2;
					double ddvm2 = (e1.vm2 - e0.vm2)/(e1.x - e0.x);
					double vm2 = e0.vm2;
					double ddwm2 = (e1.wm2 - e0.wm2)/(e1.x - e0.x);
					double wm2 = e0.wm2;

					// rounding down when casting to int
					for (int x = (int)e0.x; x < (int)e1.x; x++)
					{
						if (65535.0 - z > context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)])
						{
							Color color = shader.calculate(
								material,
								new Vector3D(um0/(z-context.PerspectiveCorrection), vm0/(z-context.PerspectiveCorrection), wm0/(z-context.PerspectiveCorrection)),
								new Vector3D(um1/(z-context.PerspectiveCorrection), vm1/(z-context.PerspectiveCorrection), wm1/(z-context.PerspectiveCorrection)),
								new Vector3D(um2/(z-context.PerspectiveCorrection), vm2/(z-context.PerspectiveCorrection), wm2/(z-context.PerspectiveCorrection)));

							context.Surface.SetPixelAlpha(x, y, context.Fog.applyFog(z/65535.0, color));

							context.DepthBuffer[(int)(x - context.Viewport.Left), (int)(y - context.Viewport.Top)] = (ushort)(65535.0 - z);
						}

						z += ddZ;

						um0 += ddum0;
						vm0 += ddvm0;
						wm0 += ddwm0;

						um1 += ddum1;
						vm1 += ddvm1;
						wm1 += ddwm1;

						um2 += ddum2;
						vm2 += ddvm2;
						wm2 += ddwm2;
					}

					e0.x += e0.dx;
					e0.z += e0.dz;

					e0.um0 += e0.dum0;
					e0.vm0 += e0.dvm0;
					e0.wm0 += e0.dwm0;

					e0.um1 += e0.dum1;
					e0.vm1 += e0.dvm1;
					e0.wm1 += e0.dwm1;

					e0.um2 += e0.dum2;
					e0.vm2 += e0.dvm2;
					e0.wm2 += e0.dwm2;

					e1.x += e1.dx;
					e1.z += e1.dz;

					e1.um0 += e1.dum0;
					e1.vm0 += e1.dvm0;
					e1.wm0 += e1.dwm0;

					e1.um1 += e1.dum1;
					e1.vm1 += e1.dvm1;
					e1.wm1 += e1.dwm1;

					e1.um2 += e1.dum2;
					e1.vm2 += e1.dvm2;
					e1.wm2 += e1.dwm2;
				}

				y++;
			}
		}

		public void Render(RenderingContext context, Material material, RenderingModes mode)
		{
			switch (mode)
			{
				case RenderingModes.WIREFRAME:
					RenderWireframe(context, material.Color);
					break;

				case RenderingModes.FLAT:
					RenderFlat(context, material);
					break;

				case RenderingModes.GOURAUD:
					RenderGouraud(context, material);
					break;

				case RenderingModes.BUMPMAP:
				{
					Matrix matrixWorld2Local = context.Local2World.Inverted;

					BumpMapVertexShader vertexShader = new BumpMapVertexShader(
						context.Camera2Canonical * context.World2Camera * context.Local2World,
						matrixWorld2Local * ((PositionalLight)context.Light).Position,
						matrixWorld2Local * context.Camera.Position);

					BumpMapPixelShader pixelShader = new BumpMapPixelShader();

					RenderShader(context, material, vertexShader, pixelShader);

					break;
				}

				case RenderingModes.SPECULAR:
				{
					Matrix matrixWorld2Local = context.Local2World.Inverted;

					SpecularVertexShader vertexShader = new SpecularVertexShader(
						context.Camera2Canonical * context.World2Camera * context.Local2World,
						matrixWorld2Local * ((PositionalLight)context.Light).Position,
						matrixWorld2Local * context.Camera.Position);

					SpecularPixelShader pixelShader = new SpecularPixelShader();

					RenderShader(context, material, vertexShader, pixelShader);

					break;
				}

				case RenderingModes.TEXTURE:
				{
					Matrix matrixWorld2Local = context.Local2World.Inverted;

					TextureVertexShader vertexShader = new TextureVertexShader(
						context.Camera2Canonical * context.World2Camera * context.Local2World);

					TexturePixelShader pixelShader = new TexturePixelShader();

					RenderShader(context, material, vertexShader, pixelShader);

					break;
				}

				case RenderingModes.DISPLAY:
				{
					TexturePixelShader pixelShader = new TexturePixelShader();

					RenderDisplay(context, material, pixelShader);

					break;
				}
			}
		}
	}
}
