using System;

namespace Sharp3D
{
	public class SpotLight : ILight
	{
		private Color color;
		private Point3D position;
		private Vector3D direction;
		private double innerCone, outerCone;
		private double attenuation;

		// 'innerCone' and 'outerCone': cos of angle
		public SpotLight(Color color, Point3D position, Vector3D direction,
			double innerCone, double outerCone, double attenuation)
		{
			this.color = color;
			this.position = position;
			this.direction = direction.GetNormalized();
			this.innerCone = innerCone;
			this.outerCone = outerCone;
			this.attenuation = attenuation;
		}

		private double getGamma(double cosPhi, double dist)
		{
			if (innerCone < cosPhi)
			{
				double t = attenuation * dist;
				return 1.0 / (t*t + 1.0);
			}
			else if (outerCone < cosPhi)
			{
				double t = attenuation * dist;
				return (cosPhi - outerCone) / ((innerCone - outerCone) * (t*t + 1.0));
			}
			else
			{
				return 0.0;
			}
		}

		public double CalcDiffuseWeight(Point3D position, Vector3D normal, Material material)
		{
			Vector3D light = this.position - position;
			double dist = light.Length;
			light /= dist;

			double sc = light * normal;

			if (sc < 0.0)
			{
				return 0.0;
			}
			else
			{
				return sc * getGamma(-light * direction, dist);
			}
		}

		public double CalcSpecularWeight(Point3D position, Vector3D normal, Point3D cameraPos, Material material)
		{
			Vector3D light = this.position - position;
			double dist = light.Length;
			light /= dist;

			Vector3D reflect = 2.0*(light*normal)*normal - light;
			Vector3D view = (cameraPos - position).GetNormalized();

			double sc = view * reflect;

			if (sc < 0.0)
			{
				return 0.0;
			}
			else
			{
				return Math.Pow(sc, material.SpecularExponent) * getGamma(-light * direction, dist);
			}
		}

		public ILight GetTransformed(Matrix m)
		{
			return new SpotLight(color, m*position, m*direction, innerCone, outerCone, attenuation);
		}

		public Color Color
		{
			get
			{
				return color;
			}
		}
	}
}
