using System;
using System.Collections;


namespace Sharp3D 
{
	public class Surface
	{
		unsafe short* bits;
		private int pitch;

		int width;
		int height;

		unsafe public Surface(short* bits, int pitch, int width, int height) 
		{
			this.bits = bits;
			this.pitch = pitch;
			this.width = width;
			this.height = height;
		}

		public int Width
		{
			get
			{
				return width;
			}
		}

		public int Height
		{
			get
			{
				return height;
			}
		}

		unsafe public void SetPixel(int x, int y, Color color)
		{
			if (x < width && y < height)	// need additional check due to avoided fringes to the right/bottom
				bits[y*pitch + x] = color.Value16;
		}

		unsafe public void SetPixelAlpha(int x, int y, Color color)
		{
			if (x < width && y < height)
			{
				double scale = color.AlphaScaled;

				bits[y * pitch + x] = ((1.0 - scale) * new Color(bits[y * pitch + x]) + scale * color).Value16;
			}
		}

		unsafe public void Clear(Color color)
		{
			short color16 = color.Value16;

			for (int i = 0; i < height*pitch; i++)
				bits[i] = color16;
		}

		public void DrawLineDDA(Line2D line, Color color)
		{
			if (Math.Abs(line.EndPoint.Y - line.StartPoint.Y) < Math.Abs(line.EndPoint.X - line.StartPoint.X))
			{
				double s = (line.EndPoint.Y - line.StartPoint.Y) / (line.EndPoint.X - line.StartPoint.X);

				if (line.StartPoint.X < line.EndPoint.X)
				{
					int x = (int)line.StartPoint.X;
					double y = line.StartPoint.Y;

					while (x <= (int)line.EndPoint.X)
					{
						SetPixel(x, (int)Math.Round(y), color);

						y += s;
						x++;
					}
				}
				else
				{
					int x = (int)line.EndPoint.X;
					double y = line.EndPoint.Y;

					while (x <= (int)line.StartPoint.X)
					{
						SetPixel(x, (int)Math.Round(y), color);

						y += s;
						x++;
					}
				}
			}
			else
			{
				double s = (line.EndPoint.X - line.StartPoint.X) / (line.EndPoint.Y - line.StartPoint.Y);

				if (line.StartPoint.Y < line.EndPoint.Y)
				{
					int y = (int)line.StartPoint.Y;
					double x = line.StartPoint.X;

					while (y <= (int)line.EndPoint.Y)
					{
						SetPixel((int)Math.Round(x), y, color);

						x += s;
						y++;
					}
				}
				else
				{
					int y = (int)line.EndPoint.Y;
					double x = line.EndPoint.X;

					while (y <= (int)line.StartPoint.Y)
					{
						SetPixel((int)Math.Round(x), y, color);

						x += s;
						y++;
					}
				}
			}
		}

		public void DrawLineBresenham(Line2D line, Color color)
		{
			if (Math.Abs(line.EndPoint.Y - line.StartPoint.Y) < Math.Abs(line.EndPoint.X - line.StartPoint.X))
			{
				if (line.EndPoint.X < line.StartPoint.X)
				{
					// swap points

					Point2D tmp = line.EndPoint;
					line.EndPoint = line.StartPoint;
					line.StartPoint = tmp;
				}

				int dx = (int)(line.EndPoint.X - line.StartPoint.X);
				int dy = (int)(line.EndPoint.Y - line.StartPoint.Y);

				// s == 2 * dy
				int f = 2 * dy;

				int x = (int)line.StartPoint.X;
				int y = (int)line.StartPoint.Y;

				int endX = (int)line.EndPoint.X;

				if (dy > 0)
				{
					while (x <= endX)
					{
						SetPixel(x, y, color);

						x++;
						if (f >= dx)
						{
							y++;
							f -= 2 * dx;
						}
						f += 2 * dy;
					}
				}
				else
				{
					while (x <= endX)
					{
						SetPixel(x, y, color);

						x++;
						if (f <= -dx)
						{
							y--;
							f += 2 * dx;
						}
						f += 2 * dy;
					}
				}
			}
			else
			{
				if (line.EndPoint.Y < line.StartPoint.Y)
				{
					// swap points

					Point2D tmp = line.EndPoint;
					line.EndPoint = line.StartPoint;
					line.StartPoint = tmp;
				}

				int dx = (int)(line.EndPoint.X - line.StartPoint.X);
				int dy = (int)(line.EndPoint.Y - line.StartPoint.Y);

				// s == 2 * dx
				int f = 2 * dx;

				int x = (int)line.StartPoint.X;
				int y = (int)line.StartPoint.Y;

				int endY = (int)line.EndPoint.Y;

				if (dx > 0)
				{
					while (y <= endY)
					{
						SetPixel(x, y, color);

						if (f >= dy)
						{
							x++;
							f -= 2 * dy;
						}
						f += 2 * dx;

						y++;
					}
				}
				else
				{
					while (y <= endY)
					{
						SetPixel(x, y, color);

						if (f <= -dy)
						{
							x--;
							f += 2 * dy;
						}
						f += 2 * dx;

						y++;
					}
				}
			}
		}

		public void DrawPolygon(Polygon2D polygon, Color color)
		{
			// clip polygon

			ClipperRect2D clipper = new ClipperRect2D(new Rectangle2D(0.0, 0.0, width, height));
			Polygon2D clipped = new Polygon2D();

			if (clipper.Clip(polygon, ref clipped))
			{
				// draw it

				for (int i = 0; i < clipped.Size; i++)
				{
					DrawLineBresenham(new Line2D(clipped[i], clipped[(i+1)%clipped.Size]), color);
				}
			}
		}

		private class Edge : IComparable
		{
			public int ymin, ymax;
			public double x, deltax;

			public Edge(int ymin, int ymax, double x, double deltax)
			{
				this.ymin = ymin;
				this.ymax = ymax;
				this.x = x;
				this.deltax = deltax;
			}

			public int CompareTo(object obj)
			{
				if (x < ((Edge)obj).x)
				{
					return -1;
				}
				else if (x == ((Edge)obj).x)
				{
					return 0;
				}
				else
				{
					return 1;
				}
			}
		}

		public void FillPolygon(Polygon2D polygon, Color color)
		{
			// clip polygon

			ClipperRect2D clipper = new ClipperRect2D(new Rectangle2D(0.0, 0.0, width, height));
			Polygon2D clipped = new Polygon2D();

			if (clipper.Clip(polygon, ref clipped))
			{
				// create edge table

				ArrayList ET = new ArrayList();
				ArrayList AET = new ArrayList();

				double ymin = height;

				for (int i = 0; i < clipped.Size; i++)
				{
					double y0 = clipped[i].Y;
					double y1 = clipped[(i+1)%clipped.Size].Y;

					if (y0 != y1)
					{
						double x0 = clipped[i].X;
						double x1 = clipped[(i+1)%clipped.Size].X;

						if (y0 < y1)
						{
							ET.Add(new Edge((int)y0, (int)y1, x0, (x1 - x0)/(y1 - y0)));

							if (y0 < ymin)
								ymin = y0;
						}
						else
						{
							ET.Add(new Edge((int)y1, (int)y0, x1, (x1 - x0)/(y1 - y0)));

							if (y1 < ymin)
								ymin = y1;
						}
					}
				}

				// draw pixels

				int y = (int)ymin;
				while (ET.Count != 0 || AET.Count != 0)
				{
					int i = 0;
					while (i != ET.Count)
					{
						if (((Edge)ET[i]).ymin == y)
						{
							AET.Add(ET[i]);
							ET.RemoveAt(i);
						}
						else
						{
							i++;
						}
					}

					i = 0;
					while (i != AET.Count)
					{
						if (((Edge)AET[i]).ymax == y)
						{
							AET.RemoveAt(i);
						}
						else
						{
							i++;
						}
					}

					if (AET.Count % 2 == 1)
						throw new ApplicationException("must be even number of corners");

					AET.Sort();
					for (i = 0; i < AET.Count/2; i++)
					{
						Edge e0 = (Edge)AET[2*i];
						Edge e1 = (Edge)AET[2*i+1];

						// rounding down when casting to int
						for (int x = (int)e0.x; x < (int)e1.x; x++)
							SetPixel(x, y, color);

						e0.x += e0.deltax;
						e1.x += e1.deltax;
					}

					y++;
				}
			}
		}

		unsafe public void SaveImage(string filename)
		{
			System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(width, height);

			for (int y = 0; y < height; y++)
			{
				for (int x = 0;  x < width; x++)
				{
					Sharp3D.Color color = new Color(bits[y*pitch + x]);
					bitmap.SetPixel(x, y, System.Drawing.Color.FromArgb(color.Red, color.Green, color.Blue));
				}
			}

			bitmap.Save(filename, System.Drawing.Imaging.ImageFormat.Bmp);
		}
	}
}