using System;

namespace Sharp3D
{
	public class VectorException : System.ApplicationException
	{
		public VectorException(string message) : base(message) {
		}
	}

	public class Vector3D
	{
		private const double EPSILON = 1e-5;

		public double X,Y,Z;

		public Vector3D()
		{
			X = 0.0;
			Y = 0.0;
			Z = 0.0;
		}

		public Vector3D(double x, double y, double z)
		{
			X = x;
			Y = y;
			Z = z;
		}

		public Vector3D(Vector3D vector)
		{
			X = vector.X;
			Y = vector.Y;
			Z = vector.Z;
		}

		public Vector3D Normalize()
		{
			Length = 1.0;
			return this;
		}

		public Vector3D GetNormalized()
		{
			double length = Length;

			if (length == 0.0)
				throw new VectorException("'Normalize()' on zero vector called");

			return new Vector3D(X/length, Y/length, Z/length);
		}

		public double Length
		{
			get
			{
				return Math.Sqrt(X*X + Y*Y + Z*Z);
			}
			set
			{
				double length = Length;

				if (length == 0.0)
					throw new VectorException("'Normalize()' on zero vector called");

				X *= value/length;
				Y *= value/length;
				Z *= value/length;
			}
		}

		public double GetAngle(Vector3D vector)
		{
			double cosPhi = (this*vector) / (Length*vector.Length);

			// handle numerical errors
			if (cosPhi > 1.0)
			{
				return 0.0;
			}
			else if (cosPhi < -1.0)
			{
				return Math.PI;
			}
			else
			{
				return Math.Acos(cosPhi);
			}
		}

		public bool IsEpsilonOrthogonal(Vector3D vector)
		{
			return Math.Abs(GetAngle(vector) - 0.5*Math.PI) < EPSILON;
		}

		public static double operator* (Vector3D vectorL, Vector3D vectorR)
		{
			return vectorL.X*vectorR.X + vectorL.Y*vectorR.Y + vectorL.Z*vectorR.Z;
		}

		public bool IsEpsilonNull()
		{
			return Length < EPSILON;
		}

		public static Vector3D operator% (Vector3D vectorL, Vector3D vectorR)
		{
			return new Vector3D(
				vectorL.Y*vectorR.Z - vectorL.Z*vectorR.Y,
				vectorL.Z*vectorR.X - vectorL.X*vectorR.Z,
				vectorL.X*vectorR.Y - vectorL.Y*vectorR.X);
		}

		public Vector3D CreateOrthogonalVector()
		{
			if (Length == 0.0)
				throw new VectorException("'CreateOrthogonalVector()' on zero vector called");

			// make sure we don't return null vector

			if (Z != 0.0)
				return new Vector3D(Z, 0.0, -X).Normalize();
			else
				return new Vector3D(-Y, X, 0.0).Normalize();
		}

		public static Vector3D operator* (double val, Vector3D vector)
		{
			return new Vector3D(val*vector.X, val*vector.Y, val*vector.Z);
		}

		public static Vector3D operator* (Vector3D vector, double val)
		{
			return new Vector3D(val*vector.X, val*vector.Y, val*vector.Z);
		}

		public static Vector3D operator/ (Vector3D vector, double val)
		{
			return (1.0/val) * vector;
		}

		public static Vector3D operator- (Vector3D vector)
		{
			return new Vector3D(-vector.X, -vector.Y, -vector.Z);
		}

		public static Vector3D operator+ (Vector3D vectorL, Vector3D vectorR)
		{
			return new Vector3D(vectorL.X + vectorR.X, vectorL.Y + vectorR.Y, vectorL.Z + vectorR.Z);
		}

		public static Vector3D operator - (Vector3D vectorL, Vector3D vectorR)
		{
			return new Vector3D(vectorL.X - vectorR.X, vectorL.Y - vectorR.Y, vectorL.Z - vectorR.Z);
		}
	}
}
